module udp_wrapper (
	input				clk,
	input				reset,
	
	output wire			enet_clk,
	input				enet_int,
	output wire			enet_rst_n,
	output wire			enet_cs_n,
	output wire			enet_cmd,
	output wire			enet_wr_n,
	output wire			enet_rd_n,
	inout wire [15:0]	enet_data,
	
	input				udp_clk,
	input				udp_wr,
	input wire [15:0]	udp_data,
    output wire			udp_full,
    
    input				halt
    );
    
    /*** START MODULE INSTANTIATIONS ***/
    
    /*** Ethernet controller ***/
    (* keep *) wire				controller_full;
	(* keep *) wire				controller_status;
	(* keep *) wire				controller_wr;
	(* preserve *) reg			controller_halt, idle;
	(* preserve *) reg [15:0]	controller_data;
	
	assign controller_wr = !controller_full & !idle;
	
	dm9000a_controller ethernet (.clk(clk),
								.reset(reset),
								
								.enet_clk(enet_clk),
								.enet_int(enet_int),
								.enet_rst_n(enet_rst_n),
								.enet_cs_n(enet_cs_n),
								.enet_cmd(enet_cmd),
								.enet_wr_n(enet_wr_n),
								.enet_rd_n(enet_rd_n),
								.enet_data(enet_data),
								
								.rx_fifo_rd_clk(clk),
								.rx_fifo_rd_req(1'b1),
								.rx_fifo_rd_data(),
								.rx_fifo_empty(),
								
								.tx_fifo_wr_clk(clk),
								.tx_fifo_wr_req(controller_wr),
								.tx_fifo_wr_data(controller_data),
								.tx_fifo_full(controller_full),
								
								.halt(1'b0),
								.link_status(controller_status)
								);
	
	/*** FIFO from outside ***/
	(*preserve*)reg			udp_rd;
	(*keep*)wire		udp_empty, udp_rd_actual;
	(*keep*)wire [15:0]	udp_rd_data;
	
	assign udp_rd_actual = udp_rd & !controller_full;
	
	rx_data_fifo #(.DEPTH(16)) fifo(
		.sclr(reset),
		
		.wrclk(udp_clk),
		.data(udp_data),
		.wrreq(udp_wr),
		.wrfull(udp_full),
		
		.rdclk(clk),
		.rdreq(udp_rd_actual),
		.q(udp_rd_data),
		.rdempty(udp_empty)
		);
	
	/*** Intermediate stuff ***/

	(*preserve*)reg [15:0]		data;
	(*preserve*)reg [10:0]		data_length;
	(*keep*)wire [10:0]		ip_length;
	(*keep*)wire [10:0]		udp_length;
	
	assign ip_length = data_length + 11'd28;
	assign udp_length = data_length + 11'd8;
								
	(*preserve*)reg [4:0]		state;
	reg [16:0]		checksum;
	wire [15:0]		checksumsum;
	
	assign checksumsum = checksum[15:0] + checksum[16];

	always @ (posedge clk) begin
		if (reset) begin
			state <= 5'd30;
			
			udp_rd <= 1'b0;
			
			idle <= 1'b1;
			controller_data <= 16'h0000;
			
			data <= 16'h0001;
			data_length <= 16'b0;
			
			checksum <= 16'b0;
		end
		else if (controller_full) begin
			state <= state;
			controller_data <= controller_data;
		end
		else begin
			case (state)
				5'd30:	begin // Spin stuff
							state <= udp_empty ? state : 5'd0;
							idle <= 1'b1;
							data_length <= udp_empty ? 11'b0 : udp_rd_data;
							udp_rd <= udp_empty ? 1'b0 : 1'b1;
							checksum <= 16'b0;
						end
				5'd0:	begin // Data transfer size
							state <= 5'd1;
							idle <= 1'b0;
							controller_data <= ip_length;
							udp_rd <= 1'b0;
						end
				5'd1:	begin // Start destination MAC address
							state <= 5'd2;
							controller_data <= 16'hFFFF;
							data_length <= data_length;
						end
				5'd2:	begin
							state <= 5'd3;
							controller_data <= 16'hFFFF;
							data_length <= data_length;
						end
				5'd3:	begin // Last destination MAC address
							state <= 5'd4;
							controller_data <= 16'hFFFF;
							data_length <= data_length;
						end
				5'd4:	begin // EtherType
							state <= 5'd5;
							controller_data <= 16'h0008;
							data_length <= data_length;
						end
				5'd5:	begin // IP Frame Start
							state <= 5'd6;
							controller_data <= {8'd0, 8'h45};
							checksum <= checksumsum + {16'h4500};
							data_length <= data_length;
						end
				5'd6:	begin
							state <= 5'd7;
							controller_data <= {ip_length[7:0], 5'b0, ip_length[10:8]};
							checksum <= checksumsum + ip_length;
							data_length <= data_length;
						end
				5'd7:	begin
							state <= 5'd8;
							controller_data <= 16'h0000;
							checksum <= checksumsum;
							data_length <= data_length;
						end
				5'd8:	begin
							state <= 5'd9;
							controller_data <= 16'h0000;
							checksum <= checksumsum;
							data_length <= data_length;
						end
				5'd9:	begin
							state <= 5'd10;
							controller_data <= {8'd17, 8'd20};
							checksum <= checksumsum + {8'd20, 8'd17};
							data_length <= data_length;
						end
				5'd10:	begin
							state <= 5'd11;
							controller_data <= ~{checksumsum[7:0], checksumsum[15:8]};
							data_length <= data_length;
						end
				5'd11:	begin
							state <= 5'd12;
							controller_data <= 16'hFFFF;
							data_length <= data_length;
						end
				5'd12:	begin
							state <= 5'd13;
							controller_data <= 16'hFFFF;
							data_length <= data_length;
						end
				5'd13:	begin
							state <= 5'd14;
							controller_data <= 16'hFFFF;
							data_length <= data_length;
						end
				5'd14:	begin
							state <= 5'd17;
							controller_data <= 16'hFFFF;
							data_length <= data_length;
						end
				5'd17:	begin // Start UDP frame
							state <= 5'd18;
							controller_data <= 16'h0;
							data_length <= data_length;
						end
				5'd18:	begin
							state <= 5'd19;
							controller_data <= 16'h8D7A;
							data_length <= data_length;
						end
				5'd19:	begin
							state <= 5'd20;
							controller_data <= {udp_length[7:0], 5'b0, udp_length[10:8]};
							data_length <= data_length;
						end
				5'd20:	begin
							state <= 5'd21;
							controller_data <= 16'h0;
							data_length <= data_length;
							udp_rd <= 1'b1;
						end
				5'd21:	begin
							state <= (udp_empty) ? 5'd21 :
								(data_length <= 11'd2) ? 5'd22 : 5'd21;
							data_length <= (udp_empty) ? data_length : data_length - 11'd2;
							controller_data <= udp_rd_data;
							udp_rd <= 1'b1;
						end
				5'd22:	begin
							state <= 5'd30;
							idle <= 1'b1;
							controller_data <= 16'h0000;
							udp_rd <= 1'b0;
						end
			endcase
		end
	end

endmodule
